// Starsiege "InstantAI"
// Authors: [OF] Pincushion
// Last updated: Tue Sep 11 00:09:33 PDT 2001
//
// See param.cs for parameters, iai.html for documentation.

$IAI::Version = "1.0";
echo("InstantAI "@$IAI::Version);

deleteFunctions("IAI::*");

$IAI::Folder = "iai";

exec($IAI::Folder@"\\param.cs");

function IAI::Msg(%msg) {
	%msg = "IAI: "@%msg;
	echo(%msg);
	//missionLog(%msg);
	//say(0,0,"<F1>"@%msg);
}

function IAI::IsDisabled() {
	if ($me::enableMissionEditor) {
		echo("Disabled while using the mission editor.");
		return true;
	}
	if ($IAI::TeamSize <= 0) {
		echo("Disabled ($IAI::Teamsize="@$IAI::TeamSize@").");
		return true;
	}
	if ($server::MissionFileName == "MapVote.mis") {
		echo("Disabled during MapVote.");
		return true;
	}
	return false;
}

function IAI::IsArty(%vid) {
	%s = getVehicleName(%vid);
	if (%s == *IDVEH_TR_NIKE) return true; // Nike
	if (%s == *IDVEH_RB_ARTL) return true; // (Rebel) Artillery
	if (%s == *IDVEH_CY_ARTL) return true; // (Cybrid) Artillery
	return false;
}

// %vid is a player or AI id
function IAI::NameID(%vid) {
	if (%vid) %name = getHudName(%vid); else %name = "NULL";
	%name = %name@"("@getObjectName(%vid)@")";
	return %name@"("@%vid@")";
}

function IAI::SetVehicleFile (%fn) {
	$IAI::VehicleFile = %fn;
	$IAI::VehN = 0;
	exec($IAI::Folder@"\\v\\"@%fn@".cs");
}

function IAI::Initialize() {
	if (IAI::IsDisabled()) return(0);
	if (getObjectId("MissionGroup/AIGroup")) return(0);

	%id = newObject("AIGroup",SimGroup);
	addToSet("MissionGroup",%id);
	$IAI::AIGroup = %id;

	$IAI::TID2Team[0] = *IDSTR_TEAM_YELLOW;
	$IAI::TID2Team[1] = *IDSTR_TEAM_BLUE;
	$IAI::TID2Team[2] = *IDSTR_TEAM_RED;
	$IAI::TID2Team[3] = *IDSTR_TEAM_PURPLE;
	$IAI::Team2TID[*IDSTR_TEAM_YELLOW] = 0;
	$IAI::Team2TID[*IDSTR_TEAM_BLUE] = 1;
	$IAI::Team2TID[*IDSTR_TEAM_RED] = 2;
	$IAI::Team2TID[*IDSTR_TEAM_PURPLE] = 3;
	$IAI::TID2Color[0] = "yellow";
	$IAI::TID2Color[1] = "blue";
	$IAI::TID2Color[2] = "red";
	$IAI::TID2Color[3] = "purple";

	// CCC = chat color code
	$IAI::TID2CCC[0] = "<F3>";
	$IAI::TID2CCC[1] = "<F1>";
	$IAI::TID2CCC[2] = "<F5>";
	$IAI::TID2CCC[3] = "<F2>";

	if ($IAI::GameType == "") {
		// no GameType specified, determine what it is based on
		// whether CTFStdlib.cs has been called
		if (dataRetrieve(0,"blueFlagCount"))
			$IAI::GameType = "CTF";
		else
			$IAI::GameType = "TDM";
	}

	$IAI::AIcnt = 0;

	// Max players per team. If necessary AI will be added to make this true.
	for (%tid=0;%tid<4;%tid++) $IAI::MaxPlayers[%tid] = $IAI::TeamSize;

	if (!$server::AllowTeamYellow) $IAI::MaxPlayers[0] = 0;
	if (!$server::AllowTeamBlue) $IAI::MaxPlayers[1] = 0;
	if (!$server::AllowTeamRed) $IAI::MaxPlayers[2] = 0;
	if (!$server::AllowTeamPurple) $IAI::MaxPlayers[3] = 0;

	$IAI::NTeams = 0;
	for (%tid=0;%tid<4;%tid++) if ($IAI::MaxPlayers[%tid]) $IAI::NTeams++;

	exec($IAI::Folder@"\\p\\"@$IAI::PilotFile@".cs");
	for (%i=$IAI::MinPilot;%i<=$IAI::MaxPilot;%i++)
		$IAI::PilotVID[%i] = 0;
	IAI::Msg("Created "@$IAI::MaxPilot-$IAI::MinPilot@" pilots.");

	IAI::SetVehicleFile($IAI::VehicleFile);

	if ($IAI::ModifyHostname) {
		if ($IAI::Hostname == "")
			$IAI::Hostname = $server::Hostname;
		%n = $IAI::TeamSize * $IAI::NTeams;
		$server::Hostname = $IAI::Hostname@" IAI";
		if (%n) $server::Hostname = $server::Hostname@%n;
	}

	IAI::MakePaths();

	IAI::Msg("Ready.");
	return(1);
}

// Counts human players per team.
function IAI::CountHumans() {
	for (%tid=0;%tid<4;%tid++) $IAI::NHuman[%tid] = 0;
	%n = playerManager::GetPlayerCount();
	for (%i=0;%i<%n;%i++) {
		%p = playerManager::GetPlayerNum(%i);
		%t = getTeam(%p);
		$IAI::NHuman[$IAI::Team2TID[%t]]++;
		%vid = playerManager::playerNumToVehicleId(%p);
		if ($IAI::HumanHealTrickle)
			healObject(%vid,$IAI::HumanHealTrickle);
		if ($IAI::HumanReloadTrickle)
			reloadObject(%vid,$IAI::HumanReloadTrickle);
	}
}

// Counts AI players per team.
function IAI::CountAI() {
	for (%tid=0;%tid<4;%tid++) {
		$IAI::NAI[%tid] = 0;
		$IAI::NArty[%tid] = 0;
	}
	%vid = 0;
	while (%vid=getNextObject($IAI::AIGroup,%vid)) {
		%tid = $IAI::Team2TID[getTeam(%vid)];
		$IAI::NAI[%tid]++;
		if ($IAI::ArtyPerTeam)
			if (IAI::IsArty(%vid)) $IAI::NArty[%tid]++;
		if ($IAI::AIHealTrickle)
			healObject(%vid,$IAI::AIHealTrickle);
		if ($IAI::AIReloadTrickle)
			reloadObject(%vid,$IAI::AIReloadTrickle);
	}
}

function IAI::CountPlayers() {
	IAI::CountHumans();
	IAI::CountAI();
}

function IAI::RandomHuman(%tid) {
	%r = randomInt(1,$IAI::NHuman[%tid]);
	%cnt = 0;
	%n = playerManager::GetPlayerCount();
	for (%i=0;%i<%n;%i++) {
		%pid = playerManager::GetPlayerNum(%i);
		if (%tid == $IAI::Team2TID[getTeam(%pid)]) {
			%cnt++;
			%vid = playerManager::playerNumToVehicleId(%pid);
			if (%cnt == %r) return %vid;
		}
	}
	IAI::Msg("BUG RandomHuman");
	return 0;
}

function IAI::RandomAI(%tid) {
	if ($IAI::NAI[%tid] < 2) return; // nobody else to guard
	%r = randomInt(1,$IAI::NAI[%tid]);
	%cnt = 0;
	%vid = 0;
	while (%vid=getNextObject($IAI::AIGroup,%vid)) {
		if (%tid == $IAI::Team2TID[getTeam(%vid)]) {
			%cnt++;
			if (%cnt == %r) return %vid;
		}
	}
	IAI::Msg("BUG RandomAI");
	return 0;
}

function IAI::AccumulateCenter(%gname) {
	%gid = getObjectId(%gname);
	%id = 0;
	while (%id=getNextObject(%gid,%id)) {
		$IAI::CenterX = $IAI::CenterX + getPosition(%id,x);
		$IAI::CenterY = $IAI::CenterY + getPosition(%id,y);
		$IAI::CenterN++;
	}
}

function IAI::MakeSquarePath(%group,%x,%y,%z,%radius) {
	%gid = newObject(%group,SimGroup);
	addToSet("MissionGroup",%gid);
	%xgroup = "MissionGroup/"@%group;
	%fn = $IAI::Folder@"/obj/PathMarker.dat";
	%id = loadObject("P1",%fn);
	setPosition(%id,%x-%radius,%y-%radius,%z);
	addToSet(%xgroup,%id);
	%id = loadObject("P2",%fn);
	setPosition(%id,%x-%radius,%y+%radius,%z);
	addToSet(%xgroup,%id);
	%id = loadObject("P3",%fn);
	setPosition(%id,%x+%radius,%y+%radius,%z);
	addToSet(%xgroup,%id);
	%id = loadObject("P4",%fn);
	setPosition(%id,%x+%radius,%y-%radius,%z);
	addToSet(%xgroup,%id);
	if ($IAI::PathBeam) {
		%id = newObject(%group@"Beam",StaticShape,"hTitanBEAM.DTS");
		addToSet("MissionGroup",%id);
		setPosition(%id,%x,%y,%z-100);
	}
}

function IAI::MakePaths() {
	$IAI::CenterX = 0;
	$IAI::CenterY = 0;
	$IAI::CenterN = 0;
	if (IAI::IsDisabled()) return;

	if($server::TeamPlay == true) {
		for (%tid=0;%tid<4;%tid++)
			if ($IAI::MaxPlayers[%tid])
				IAI::AccumulateCenter("MissionGroup/DropPointGroup"@%tid+1);
	} else {
		IAI::AccumulateCenter("MissionGroup/DropPointGroup");
	}
	%n = $IAI::CenterN;
	if (%n == 0) {
		IAI::Msg("No DropPointGroup objects?");
		return;
	}
	%x = $IAI::CenterX / %n;
	$IAI::CenterX = %x;
	%y = $IAI::CenterY / %n;
	$IAI::CenterY = %y;
	%z = getTerrainHeight(%x,%y);
	$IAI::CenterZ = %z;

	if ($IAI::GameType == "CTF") {
		for (%tid=0;%tid<4;%tid++) {
			if ($IAI::MaxPlayers[%tid]) {
				%color = $IAI::TID2Color[%tid];
				%xnm = "MissionGroup/"@%color@"Base/"@%color@"Flag";
				%xid = getObjectId(%xnm);
				if (%xid) {
					%x = getPosition(%xid,x);
					%y = getPosition(%xid,y);
					%z = getTerrainHeight(%x,%y);
					IAI::MakeSquarePath("IAIPath"@%tid,%x,%y,%z,60);
				} else
					IAI::Msg("No "@%xnm@"?");
			}
		}
	} else {
		IAI::MakeSquarePath("IAICenterPath",%x,%y,%z,60);
	}
}

function IAI::OrderAI(%vid) {
	order(%vid,clear,true);
	order(%vid,speed,high);
	%xid = 0;
	if ($IAI::GameType == "CTF") {
		// in CTF guard the flag
		%tid = $IAI::Team2TID[getTeam(%vid)];
		%xnm = "MissionGroup/IAIPath"@%tid;
		%xid = getObjectId(%xnm);
	} else {
		// otherwise guard center point between bases of active teams
		%xnm = "MissionGroup/IAICenterPath";
		%xid = getObjectId(%xnm);
	}
if (0) {
	%tid = $IAI::Team2TID[getTeam(%vid)];
	if ($IAI::NHuman[%tid]) {
		// guard a human if any
		%xid = IAI::RandomHuman(%tid);
		%xnm = IAI::NameID(%xid);
	} else {
		%xid = IAI::RandomAI(%tid);
		%xnm = IAI::NameID(%xid);
	}
}
	if (%xid) {
		order(%vid,guard,%xid);
		//IAI::Msg("Order "@IAI::NameID(%vid)@" Guard "@%xnm);
	} else {
		IAI::Msg("No orders for "@IAI::NameID(%vid)@"?");
	}
}

function IAI::PilotRelease(%vid) {
	%pilot = dataRetrieve(%vid,Pilot);
	dataRelease(%vid,Pilot);
	//IAI::Msg("PilotRelease "@%pilot@" - "@getHUDName(%vid));
	$IAI::PilotVID[%pilot] = 0;
}

function IAI::PilotAssign(%vid) {
	for (%pilot=$IAI::MinPilot;%pilot<=$IAI::MaxPilot;%pilot++) {
		if ($IAI::PilotVID[%pilot] == 0) {
			dataStore(%vid,Pilot,%pilot);
			$IAI::PilotVID[%pilot] = %vid;
			return(%pilot);
		}
	}
	IAI::Msg("BUG PilotAssign");
	return($IAI::MinPilot);
}

function IAI::RandomVehicle(%fn) {
	if ($IAI::VehN <= 0) {
		// load AI vehicle config list if necessary
		exec($IAI::Folder@"\\v\\"@%fn@".cs");
		IAI::Msg(%fn@" specifies "@$IAI::VehN@" AI vehicle configurations.");
		IAI::StatDump(); // show last mission's stats
		IAI::StatClear();
	}
	%vix = randomInt(1,$IAI::VehN);
	%ofn = $IAI::Folder@"/obj/"@$IAI::Veh[%vix]@".dat";
if (0) {
//FIXME
	// I think loadObject() might leak,
	// this is an untested attempt to minimize calls to it
	%vid = $IAILoadObject[%ofn];
	if (%vid == "") {
		%vid = loadObject(%objname,%ofn);
		$IAILoadObject[%ofn] = %vid;
	}
	%vid = cloneVehicle(%vid);
} else {
	%vid = loadObject(%objname,%ofn);
}
	if (%vid==0)
		IAI::Msg("Object "@%vix@", "@%ofn@" FAILED.");
	else
		IAI::StatSpawn(%vid,%vix);
	return %vid;
}

function IAI::SpawnAI(%tid) {
//	IAI::Msg("SpawnAI "@$IAI::TID2Team[%tid]);
	%objname = "AI"@$IAI::AIcnt;
	if ($IAI::NArty[%tid] < $IAI::ArtyPerTeam)
		%vid = IAI::ArtyCreate(%objname);
	else
		%vid = IAI::VehicleCreate(%objname);
	if (%vid == 0) {
		IAI::Msg("Bad object!");
		return;
	}
	$IAI::AIcnt++;
	%pilot = IAI::PilotAssign(%vid);
	addToSet("MissionGroup/AIGroup",%vid);
	setTeam(%vid,$IAI::TID2Team[%tid]);
	schedule("IAI::OrderAI("@%vid@");",1);
	setPilotId(%vid,%pilot);
	redrop(%vid);
//	IAI::Msg("AI "@$IAI::TID2Team[%tid]@" "@%objname@" "@%fn);
}

// Maintain specified number of players per team.
function IAI::Daemon() {
	if (IAI::IsDisabled()) return;
	IAI::Initialize();
	IAI::CountPlayers();
if (0) {
	%msg = "Daemon "@getCurrentTime();
	for (%tid=0;%tid<4;%tid++)
		%msg = %msg @ " " @ $IAI::NHuman[%tid] @ ":" @ $IAI::NAI[%tid];
	IAI::Msg(%msg);
}
	for (%tid=0;%tid<4;%tid++)
		if (($IAI::NHuman[%tid]+$IAI::NAI[%tid]) < $IAI::MaxPlayers[%tid])
			schedule("IAI::SpawnAI("@%tid@");",%tid+1);
	schedule("IAI::Daemon();",5);
}

// vehicle::onAdd() from DMStdlib.cs, no mods except name
function IAI::DMvehicleonAdd(%this)
{
   if($server::TeamPlay != true)
   {
   	// see if it is a player
      %player = playerManager::vehicleIdToPlayerNum(%this);
      if(%player == 0)  // that's the server
         return;
         
      // so it's a player
      
      // shouldn't have to do this, but oh well
      setTeam(%this, *IDSTR_TEAM_RED);
   }
}

// vehicle::onAdd() from CTFStdlib.cs, getTeam() added and name changed.
function IAI::CTFvehicleonAdd(%this)
{
	//%team = getTeam(playerManager::vehicleIdToPlayerNum(%this));
%team = getTeam(%this);
	%color = teamToColor(%team);
	%flagKey = strcat(%color, "FlagCount");
	
   	adjTeamCount(%team, 1);
	
	//if the flag isn't at the base, but no one has it, correct situation
	if(dataRetrieve(0, %flagKey) && !dataRetrieve(0, strcat(%color, "FlagCarried")))
	{
		setFlag(%team, true);
	}
}

function vehicle::onAdd(%this) {
	%player = playerManager::vehicleIdToPlayerNum(%this);
	if (%player) {
		if (IAI::Initialize()) {
			// first human player starts daemon
			IAI::Daemon();
		}
	}
	// call original gametype-specific handler
	if ($IAI::GameType == "CTF") IAI::CTFvehicleonAdd(%this);
	if ($IAI::GameType == "TDM") IAI::DMvehicleonAdd(%this);
	if ($IAI::GameType == "DM") IAI::DMvehicleonAdd(%this);
}

function IAI::getFancyDeathMessage(%ed,%er) {
if (0) {
	// team color display
	%ced = $IAI::TID2CCC[$IAI::Team2TID[getTeam(%ed)]];
	%cer = $IAI::TID2CCC[$IAI::Team2TID[getTeam(%er)]];
	%cun = "<F0>";
	return getFancyDeathMessage(%ced@getHUDName(%ed)@%cun,
		%cer@getHUDName(%er)@%cun);
} else {
	// normal display
	return getFancyDeathMessage(getHUDName(%ed),getHUDName(%er));
}
}

// vehicle::onDestroyed() from CTFStdlib.cs, added call to
// IAI::getFancyDeathMessage().
function IAI::CTFvehicleonDestroyed(%destroyed,%destroyer)
{
   %team = getTeam(%destroyed);
   adjTeamCount(%team, -1);
   %teamCount = getTeamPlayerCount(%team);      

   if(%teamCount < 1)
   {
	   setFlag(%team, false);					
   }

   %color = dataRetrieve(%destroyed, "hasFlag");
   %flagTeam = colorToTeam(%color);
   
   // if the destroyed vehicle has a flag, it goes back to it's base
   if (%color != "")
   {
      playerDropsFlag(%destroyed);
      
      // let everyone know this guy drops the flag
      %str =
         *IDMULT_CHAT_SURRENDER_FLAG_1 @
         getName( %destroyed )         @
         *IDMULT_CHAT_SURRENDER_FLAG_2 @ 
         %flagTeam                     @
         *IDMULT_CHAT_SURRENDER_FLAG_3;
      
      %soundFile = "";   
      if(%flagTeam == *IDSTR_TEAM_RED)
      {
         %soundFile = "red_flag_sur.wav";
      }
      else if(%flagTeam == *IDSTR_TEAM_BLUE)
      {
         %soundFile = "blue_flag_sur.wav";
      }
      else if(%flagTeam == *IDSTR_TEAM_YELLOW)
      {
         %soundFile = "yel_flag_sur.wav";
      }
      else
      {
         %soundFile = "purp_flag_sur.wav";
      }   
      
      say( 0, 0, %str, %soundFile );

            
      if($scoringFreeze == false)
      {     
         // destroyer's team gets credit for killing the carrier
         %destroyerTeam = getTeam(%destroyer);  
      
         // only give points if destroyed recovered his/her own flag
         if(%destroyerTeam == colorToTeam(%color))
         {
            %key = strcat(teamToColor(getTeam(%destroyer)), "CarriersKilled");
            dataStore(0, %key, 1 + dataRetrieve(0, %key));
            
            // player gets credit, too
            %player = playerManager::vehicleIdToPlayerNum(%destroyer);
            if(%player != 0)
            {
               %player.carriersKilled = %player.carriersKilled + 1;
            }
         }

         // echo the surrender to the console for logging
         echo(*IDSTR_CONSOLE_CTF_SURRENDER @ " " @ playerManager::vehicleIdToPlayerNum(%destroyed));                        
      }	  
   }
   else
   {
      // not a flag carrier ... is it an enemy?
      if($scoringFreeze == false)
      {
         if(getTeam(%destroyed) != getTeam(%destroyer))
         {
            %player = playerManager::vehicleIdToPlayerNum(%destroyer);
            if(%player != 0)
            {
               %player.genericKills = %player.genericKills + 1;
            }
         }
      }
   }
   
   // enforce the rules
   if(
      (getTeam(%destroyed) == getTeam(%destroyer)) &&
      (%destroyed != %destroyer)
   )
   {
      antiTeamKill(%destroyer);
   }

   vehicle::onDestroyedLog(%destroyed, %destroyer);
   
   // give the death messages...
   //%message = getFancyDeathMessage(getHUDName(%destroyed), getHUDName(%destroyer));
   %message = IAI::getFancyDeathMessage(%destroyed,%destroyer);
   if(%message != "")
   {
      say( 0, 0, %message);
   }
}

// vehicle::onDestroyed() from DMStdlib.cs, added call to
// IAI::getFancyDeathMessage().
function IAI::DMvehicleonDestroyed(%destroyed, %destroyer)
{
   // left over from missionStdLib.cs
   vehicle::onDestroyedLog(%destroyed, %destroyer);
   
   // this is weird but %destroyer isn't necessarily a vehicle
   //%message = getFancyDeathMessage(getHUDName(%destroyed), getHUDName(%destroyer));
   %message = IAI::getFancyDeathMessage(%destroyed, %destroyer);
   if(%message != "")
   {
      say( 0, 0, %message);
   }
   
   // enforce the rules
   if($server::TeamPlay == true)
   {
      if(
         (getTeam(%destroyed) == getTeam(%destroyer)) &&
         (%destroyed != %destroyer)
      )
      {
         antiTeamKill(%destroyer);
      }
   }   
}

function vehicle::onDestroyed(%ed, %er) {
	if (playerManager::vehicleIdToPlayerNum(%ed) == 0) {
		// an AI got destroyed
		schedule("deleteObject("@%ed@");",1);
		IAI::PilotRelease(%ed);
		IAI::StatDeath(%ed);
	}

	// call original gametype-specific handler
	if ($IAI::GameType == "CTF") IAI::CTFvehicleonDestroyed(%ed,%er);
	else IAI::DMvehicleonDestroyed(%ed,%er);

	if (playerManager::vehicleIdToPlayerNum(%er) == 0) {
		// destroyer is an AI
		if($server::TeamPlay && (getTeam(%ed)==getTeam(%er)) && (%ed!=%er)) {
			// team-kill by an AI
			%msg = getHUDName(%er)@": Whoops, sorry about that "@getHUDName(%ed@".");
			schedule("say(0,0,\""@%msg@"\");",6);
		}
		IAI::StatKill(%er);
	}
}

// Lets players help fix their teammates
function vehicle::onScan(%ed,%er) { 
	if (getTeam(%ed) == getTeam(%er)) {
		healObject(%ed,$IAI::ScanHeal);
		reloadObject(%ed,$IAI::ScanReload);
	}
}

function IAI::PlayerOnAdd(%pid) {
	%msg = "InstantAI "@$IAI::Version@". ";
	if ($IAI::TeamSize)
		%msg = %msg@"Teams will be filled with up to "@
 $IAI::TeamSize@" AI per team. "@
 "Players may join teams in any combination, but they cannot control AI. ";
	if ($IAI::PathBeam)
		%msg = %msg@"A beam marks the point the AI fight for. ";
	if ($IAI::AIHealTrickle)
		%msg = %msg@"AI get a trickle of health/ammo. ";
	if ($IAI::HumanHealTrickle)
		%msg = %msg@"Humans get a trickle of health/ammo. ";
	if ($IAI::ScanHeal)
		%msg = %msg@"Scan teammates to repair and reload them. ";
	if ($IAI::EnableLTADS)
		%msg = %msg@"Load an LTADS and spot opponents to call for artillery. ";
	if ($IAI::ScrambleDropPoints)
		%msg = %msg@"Drop points are scrambled. ";
	say(%pid,0,"<F1>"@%msg);
}

function player::onAdd(%pid) {
	IAI::PlayerOnAdd(%pid);
}

exec($IAI::Folder@"\\stats.cs");

if ($IAI::ArtyPerTeam) {
	exec($IAI::Folder@"\\arty.cs");
	echo($IAI::ArtyPerTeam@" artillery AI per team.");
}

$IAI::LTADSMap = $IAI::EnableLTADS;
// Add a line for any map where IAI LTADS should not be used
if ($server::MissionFileName == "CTF_City_of_Snipers.mis") $IAI::LTADSMap = 0;
if ($IAI::LTADSMap) {
	exec($IAI::Folder@"\\LTADS.cs");
	echo("LTADS enabled.");
}

if ($IAI::ScrambleDropPoints) {
	exec($IAI::Folder@"\\dpscram.cs");
	%n = IAI::DPScramble();
	echo("Scrambled "@%n@" drop points.");
}

if ($IAI::CloneConfigs) {
if (!$CmdLineServer) {
	$IAI::TeamSize = 0;
	exec($IAI::Folder@"\\clone.cs");
}
}
